//------------------------------------------------------------
// contains files for reading and playing Music Studio Files
// Created Feb 16, 2011
// by Jim Patchell....the blind hacker
//------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include "MSplay.h"

int jukebox;	/*	flag for jukebox mode	*/
char *song_ram[24];
char *song_list[24];
int	song_pointer;
int number_of_songs;

/*
	GLOBAL VARIABLES
*/

STACK *pRepeatStack;


/*
	timer globals
*/

int current_preset[16];
int track_enables[4];
long song_size;
int active_song;
int time_sig,key_sig;
int tempo;

int dur[0x15] = {
	0,4,0,6,8,9,12,16,18,24,32,36,48,64,72,96,128,144,192,0,288 };
int key_sig_tab[128];
int octave[6] = {0,-24,-12,0,12,24};
int s_octave[6][7];

int yuck_flag;
int	active_timer = FALSE;
int play_flag;
int end_flag;
int done_flag;
int done_handshake = FALSE;
char midi_str[80];	/*	midi note buffer	*/
int midi_str_len;
int ticks;

unsigned short LittleEndian(unsigned short u)
{
	union {
		unsigned short uv;
		unsigned char cv[2];
	}i,o;

	i.uv = u;
	o.cv[0] = i.cv[1];
	o.cv[1] = i.cv[0];
	return o.uv;
}

void StartSong(MSSONG *pSong)
{
	int i,j;
	short unsigned int *wptr;
	short unsigned int u,mask;
	char *SNGdata;

	InitBuffer(pSong);
	SNGdata = (char *)&pSong->SongData[MS_OFFSET_SONG_DATA];
	ticks = 100;
//	set_timer();
	pSong->Event = 0;

	yuck_flag = FALSE;
	play_flag = FALSE;
	done_flag = FALSE;
	end_flag = FALSE;
	wptr = (unsigned short int *)SNGdata;
	for(i=0;i<4;++i,++wptr)
	{
		u = *wptr;	/*	get track enable info	*/
		printf("U=%04x\n",u & 0xffff);
		u = LittleEndian(u);
		printf("U=%04x\n",u & 0xffff);
		mask = 1;
		if(i == 0)
		{
			for(j=0;j<15;++j)
			{
				pSong->trkinfo[j].EnableFlag = (u & mask)?1:0;
				mask <<= 1;	//shift mask
			}
		}
	}
	pSong->pCurSong = (char *)wptr;
}

int parse(MSSONG *pSong)
{
	//----------------------------------------
	// This function is used to parse the song
	// data so that it can be either played or
	// what ever.
	//
	// This function will continue to parse
	// untill it has decoded a complete event
	// or in other words, all of the things
	// that happen at the same time
	//
	// parameters:
	//	pSong......pointer to song descriptor
	//
	// return value
	//	returns 1 normally.  Returns 0 when
	//	song is finished.
	//----------------------------------------

	int i,t,j;
	int shortest;
	int t1,t2;
	char *midi= new char[80];	/*	buffer for midi data to play	*/
	int repeat,loop;
	REPEAT *pR;	//pointer to repeat object
	int rv = 1;
	EVENT *pEv = NULL;

	shortest = 288;
	t = 0;
	loop = 1;
	while(loop)
	{
		i = ((int)(*pSong->pCurSong++)) & 0x0ff;
		printf("MS Token = %08x\n",i);
		switch(i)	/*	switch song data	*/
		{
			case REPEAT_STOP:
				printf("Repeat Stop\n");
				if(DecRepeatCount(pR = (REPEAT *)GetStackTop(pRepeatStack)))
					pSong->pCurSong = pR->pRstart;
			case EVENT_END:
				++pSong->Event;
				if(pSong->Event == 180)
				{
					printf("Get Ready for Bomb\n");
				}
				printf("Event End = %d\n",pSong->Event);
				if(play_flag)
				{
					ticks = shortest;
//					if(t > 0)
//						midi_ws(t - 1,midi);
					timing(pSong,shortest);
					play_flag = FALSE;
					loop = 0;
					rv = 1;
				}
				break;
			case KEY_SIGNATURE:
				*pSong->pCurSong++;	//skip 0x01?
				key_sig = ((int)(*pSong->pCurSong++)) & 0x0ff;
				printf("Key Signiture = %d\n",key_sig);
//				gen_key_tab(key_sig);
				break;
			case TEMPO:
				tempo = (int)(*pSong->pCurSong++) & 0xff;
				printf("Tempo = %d\n",tempo);
				if((tempo = 12288 / (( 48 * tempo) / 60)) > 255)
					tempo = 0;
//				*((char *)0xfffa1f) = tempo;	//set the tempo
				break;
			case BAR:	//don't do much of anything
				printf("Measure Bar");
				break;
			case TIME_SIGNATURE:
				time_sig = (int)(*pSong->pCurSong++);
				printf("Time Signiture = %d\n",time_sig);
				break;
			case LOUDNESS:
				pSong->pCurSong++;	/*	points to 0x01	*/
				for(j=0;j<15;++j)
					pSong->trkinfo[j].velocity = *pSong->pCurSong;
				printf("Loudness = %d\n",*pSong->pCurSong);
				pSong->pCurSong++;	/*	points to next item	*/
				break;
			case REPEAT_START:
				repeat = (int)*pSong->pCurSong++;
				printf("Start Repeat = %d\n",repeat);
				pR = newRepeat(repeat,pSong->pCurSong);
				break;
			case END:
				printf("End of Song\n");
				end_flag = TRUE;
				loop = 0;
				rv = 0;
				break;
			default:
				/*
					if we got down here, what was found must be a note
					so, we first get the track/type,duration,accidentals
					and then the note number
					If, the note starts a tie, treat it like any other
					note, if a note is at the end of a tie, then find
					the note in the time out buffer.  If the note is
					the beginning and end of a tie, still find it
					in the timeout buffer
				*/
				if(i & MS_ENDTIE_FLAG)	/*	end tie ?	*/
				{
					t1 = (int)*pSong->pCurSong++;	/*	duration	*/
					t2 = ((int)*pSong->pCurSong++) & 0x7f;	/*	note mask off MSB (upside down note)	*/
					printf("Song Note or Rest with tie  %02x  %02x\n",t1,t2);
					if(!(i & MS_REST_FLAG))	/*	if not a rest	*/
					{
						switch(t1 & MS_KEY_MASK)
						{
							case INKEY:
								t2 = t2 + key_sig_tab[t2];
								break;
							case NATURAL:
								break;
							case SHARP:
								t2 += 1;
								break;
							case FLAT:
								t2 -= 1;
								break;
						}
					}
					if((pEv = search(t2,pSong->trkinfo[i & 0x0f].channel,pSong->pEvents,pSong->nEvents) ) == NULL)	/*	find the note in the timeout buffer 	*/
						pEv = GetFreeEvent(pSong->pEvents,pSong->nEvents);
				}
				else	//not the end of a tie
				{
					t1 = *pSong->pCurSong++;
					t2 = *pSong->pCurSong++;
					printf("Song Note or Rest  %02x  %02x\n",t1,t2);					if(!(i & MS_REST_FLAG))	//if not a rest
					{
						switch(t1 & MS_KEY_MASK)
						{
							case INKEY:
								t2 = t2 + key_sig_tab[t2];
								break;
							case NATURAL:
								break;
							case SHARP:
								t2 += 1;
								break;
							case FLAT:
								t2 -= 1;
								break;
						}
					}
					pEv = GetFreeEvent(pSong->pEvents,pSong->nEvents);
				}
				if(pEv)	/*	find free spot in note buffer	*/
				{
					pEv->Track = i & 0x0f;
					pEv->Channel = pSong->trkinfo[pEv->Track].channel;
					pEv->Velocity = pSong->trkinfo[pEv->Track].velocity;
					pEv->TieBegFlag = (i & MS_BEGTIE_FLAG)?1:0;
					pEv->TieEndFlag = (i & MS_ENDTIE_FLAG)?1:0;
					pEv->RestFlag = (i & MS_REST_FLAG)?1:0;	/*	tag as rest if need be	*/
					pEv->Duration = dur[t1 & MS_DURATION_MASK];
					pEv->Note = t2;
					if(pEv->Duration < shortest)
						shortest = pEv->Duration;
					if(t1 & MS_ACCENT_FLAG)	/*	accent	*/
					{
						pEv->Velocity += 40;	//boost velocity
						if(pEv->Velocity > 127)
							pEv->Velocity = 127;
					}
					if(!pEv->RestFlag && !pEv->TieEndFlag)
					{
						if(current_preset[pEv->Channel] != pSong->trkinfo[pEv->Track].patch)
						{
							midi[t++] = 0xc0 | pEv->Channel;	/*	patch change	*/
							midi[t++] = pSong->trkinfo[pEv->Track].patch - 1;
							current_preset[pEv->Channel] = pSong->trkinfo[pEv->Track].patch;
						}
						midi[t++] = (char)(pEv->Channel | 0x90);
						midi[t++] = (char)(pEv->Note) + pSong->trkinfo[pEv->Track].octave;
						midi[t++] = (char)(pEv->Velocity);
					}
					play_flag = TRUE;	/*	when we hit the END OF FRAME, play	*/
				}
				else
				{
					printf("ERROR\n");
					yuck_flag = TRUE;
				}
				break;
		}	/*	end of switch i	*/
	}	/*	end while	*/
	return rv;
}

EVENT *search(int note,int chan,EVENT *pEv, int MaxEvents)
{
	//---------------------------------
	// search events for a particular
	// note on a particular channel
	//
	// parameters:
	//	note....note to search for
	//	chan....on this particular channel
	//	pEv.....pointer to an array of EVENT objects
	//	MaxEvents...maximum number of events in array
	//
	// return value
	//	Returns NULL if no event was found
	//	otherwise it returns a pointer to
	//	the desired object
	//---------------------------------

	int i,loop;
	EVENT *pE;

	for(i=0,loop = 1;(i<MaxEvents)&&loop;++i)
	{
		if((pEv[i].Note == note) && !pEv[i].RestFlag && (chan == pEv[i].Channel) )
		{
			loop = 0;
			pE = &pEv[i];
		}
	}
	return(pE);
}

EVENT *GetFreeEvent(EVENT *pEv,int MaxEvents)
{
	//------------------------------
	// locate a free event
	//
	// parameters
	//	pEv.....pointer to an array of EVENT objects
	//	MaxEvents..number of EVENTS in array
	//
	// return value:
	//	return pointer to free event,
	// or NULL on error (non-free)
	//------------------------------
	int i,loop;
	EVENT *pE=NULL;

	for(i=0,loop = 1;(i<MaxEvents) && loop;++i)
	{
		if(pEv[i].EventFree == 0)
		{
			pE =&pEv[i];
			pE->EventFree = 1;	//mark it as IN use
			loop = 0;
		}
	}
	return(pE);
}

void WaitAllOff(MSSONG *pSong)
{
	//---------------------------------------
	// This function waits for all notes that
	// are playing to stop
	//--------------------------------------
	int i;
	register int shortest;
	EVENT *pEv = pSong->pEvents;

	shortest = 300;
	for(i=0;i<16;++i,++pEv)
	{
		if(pEv->EventFree)
		{
			if( (pEv->Duration > 0) && (pEv->Duration < shortest))
				shortest = pEv->Duration;
		}
	}
	if(shortest < 300)
	{
		timing(pSong,shortest);
		ticks = (long)shortest;
	}
	else
	{
		done_flag = TRUE;
	}
}

void timing(MSSONG *pSong,int n_duration)
{
	int midi_len;	/*	lenth of midi buffer string	*/
	int i;
	EVENT *pEv = pSong->pEvents;

	midi_len = 0;
	for(i=0;i<16;++i,pEv++)
	{
		if(pEv->EventFree)
		{
			if(((pEv->Duration -= n_duration) <= 0))
			{
				if(((!pEv->TieBegFlag) && (!pEv->RestFlag))
			 || ((!pEv->RestFlag) && (pEv->Duration < -48)))
				{
					midi_str[midi_len++] = (char)(pEv->Channel | 0x90);
					midi_str[midi_len++] = (char)(pEv->Note) + pSong->trkinfo[pEv->Track].octave;
					midi_str[midi_len++] = 0;	//turn off note
					pEv->EventFree = 0;	//free this Event object
				}
				if(pEv->RestFlag)
					pEv->EventFree = FALSE;	//free this Event Object
			}
		}
	}
	midi_len--;
	midi_str_len = midi_len;
}

void midi_off()
{
//	if(midi_str_len > 0)
//		midi_ws(midi_str_len,midi_str);
	midi_str_len = 0;
}

/*
	timer routines:
	This routine is called by the interrupt handler to increment the 
	local tick counter.
*/
int ticker(MSSONG *pSong)
{
	int rv = 1;
	ticks--;
	if(ticks == 0)
	{
		if(!parse(pSong)) rv = 0;
		if(end_flag && !midi_str_len)	/*	if end flag was hit during parse, do this */
			WaitAllOff(pSong);
	}
	if(ticks == 3l)
	{
		midi_off();
	}
//	if(ticks < -10)
//		ticks = 4;
	return rv;
}

void do_stop(void)
{
//	unset_timer();
}

MSSONG* LoadNextSong(void)
{
	MSSONG *pSong;

	if(!done_flag)
		pSong = NULL;
	else if(!jukebox)
		pSong = NULL;
	else
	{
		pSong = LoadFile(song_list[song_pointer],16);
		StartSong(pSong);
		song_pointer++;
		if(song_pointer > number_of_songs)
			song_pointer = 0;
	}
	return pSong;
}

int sPrintTrkInfo(char *s,MSSONG *pSong)
{
	int i = 0;
	int j;
	for(j=0;j<15;++j)
		i += sprintf(&s[i],"%2d->%s:#%d OCT=%d OCTn=%d:%s\n",j+1,pSong->trkinfo[j].name,pSong->trkinfo[j].patch,pSong->trkinfo[j].octaveNum,pSong->trkinfo[j].octave,pSong->trkinfo[j].EnableFlag?"ON":"OFF");
	return i;
}

void GetSongParameters(MSSONG *pSong)
{
	int i;
	MIDI_INFO *inf;

	inf = (MIDI_INFO *)&pSong->SongData[MS_OFFSET_MIDI_INFO];
	for(i=0;i<15;++i)
	{
		pSong->trkinfo[i].channel = ((int)((inf->chan_range[i-1]) & 0x0f) - 1) & 0x0f;
		pSong->trkinfo[i].octaveNum = (int)((inf->chan_range[i-1]) >> 4);
		pSong->trkinfo[i].octave = octave[pSong->trkinfo[i].octaveNum];
		pSong->trkinfo[i].patch = (int)(inf->patch[i-1]);
		memcpy(pSong->trkinfo[i].name,inf->Channel[i].Name,10);
	}
	strcpy(pSong->SongName,inf->name);	/*	copy song name	*/
}

void InitBuffer(MSSONG *pSong)
{
	int i;
	EVENT *pEv = pSong->pEvents;

	for(i=0;i<pSong->nEvents;++i,pEv++)
	{
		pEv->Channel = 0;
		pEv->Duration = 0;
		pEv->Note = 0;
		pEv->TieBegFlag = 0;
		pEv->TieEndFlag = 0;
		pEv->EventFree = 0;
		pEv->RestFlag = 0;
		pEv->Velocity = 0;
		current_preset[i] = -1;
	}
}

MSSONG * LoadFile(char *SongName,int nEvents)
{
	//------------------------------------
	// load a song file into a song descriptor
	// block.
	// parameters
	//	SongName.....pointer to name of song to load
	//	nEvents.....number of events to allocate
	//--------------------------------------

	struct _finddata_t	FileInfo;
	FILE *pFile;
	long FindHandle;
	int n;
	MSSONG *pSong = NULL;
	long lv,r;

	if((FindHandle = _findfirst(SongName,&FileInfo)) >= 0)
	{
		pFile = fopen(SongName,"rb");
		if(pFile == NULL)
		{
			printf("Could Not Open %s\n",SongName);
		}
		else
		{
			pSong = (MSSONG *)malloc(sizeof(MSSONG)+FileInfo.size + sizeof(EVENT) * nEvents +8);	//allocate memory for song
			memset(pSong,0,sizeof(MSSONG)+FileInfo.size + sizeof(EVENT) * nEvents + 8);	//zero out memory
			n = fread(pSong->SongData,sizeof(char),FileInfo.size,pFile);
			printf("File %s[%d] opened, read %d\n",SongName,FileInfo.size,n);
			fclose(pFile);
			_findclose(FindHandle);
			lv = (long)pSong;
			lv += sizeof(MSSONG)+FileInfo.size;
			r = lv % 4;
			lv += r;
			pSong->pEvents = (EVENT *)lv;
			pSong->nEvents = nEvents;
			printf("Event Buffer Address = %p\n",pSong->pEvents);
		}
	}
	active_song = TRUE;
	return pSong;
}

//------------------------------
// REPEAT functions
//------------------------------

REPEAT *newRepeat(int cnt, char *pRstrt)
{
	//----------------------------------------
	// This function creates a new
	// REPEAT object
	// parameters:
	//	cnt......number of times to repeat
	//	pRstrt..Start of repeated song data
	// 
	// return value:
	//	returns NULL on failure
	//----------------------------------------
	REPEAT *pR;

	pR = (REPEAT *)malloc(sizeof(REPEAT));
	if(pR)
	{
		pR->Count = cnt;
		pR->pRstart = pRstrt;
	}
	return pR;
}

void deleteRepeat(REPEAT *pR)
{
	free(pR);
}

int DecRepeatCount(REPEAT *pR)
{
	//-------------------------------
	// This function decrements repeat
	// count and then checks to see
	// if it is zero. 
	//
	// paramenters:
	//	pR.....ppointer to repeat object to decrement
	//
	// returns value:
	//	returns false if Count is not zero
	// returns true if count is zero
	//------------------------------
	int rv = 0;

	pR->Count--;
	if(pR->Count == 0)
		rv = 1;
	return rv;
}

//-----------------
// stack functions
//-----------------

STACK *newStack(int size)
{
	//---------------------------------------
	// Create a new stack object
	//
	// parameters
	//	size.....number of elements in stack
	//
	// return value
	//	returns a pointer to the stack object
	// on success.  Returns NULL on fail.
	//---------------------------------------
	STACK *pS;

	pS = (STACK *)malloc(sizeof(STACK) + sizeof(void *) * size);
	if(pS)
	{
		memset(pS,0,sizeof(STACK) + sizeof(void *) * size);
		pS->Size = size;
	}
	return pS;
}

void *GetStackTop(STACK *pS)
{
	//---------------------------------------
	// Gets a pointer to the top object on the
	// stack.
	//
	// parameters
	//	pS.....pointer of stack object to read
	//
	// return value
	//	returns NULL if the stack is empty
	//---------------------------------------
	void *pE = NULL;
	if(pS->nElem)
		pE = pS->data[pS->stactpointer-1];	//get top element
	return pS;
}

void *PopStack(STACK *pS)
{
	//----------------------------------------
	// This function gets the value off the
	// top of the stack and removes it
	//
	// parameters:
	//	pS.....stack to pop object from
	//
	// Return Value
	//	returns NULL if the stack was empty
	//----------------------------------------
	void *pE = NULL;
	if(pS->nElem)
	{
		pE = pS->data[--pS->stactpointer];	//get top element
		pS->nElem--;	//decrement number of elements
	}
	return pS;
}

int PushStack(STACK *pS,void *pE)
{
	//----------------------------------------
	// Push an object onto the stack
	//
	// parameters:
	//	pS.....stack to push object onto
	//	pE....pointer to object to push on stack
	//
	// return value
	//	returns 0 on success, returns 1 if 
	// stack was full
	//-----------------------------------------

	int rv = 0;

	if(pS->nElem == pS->Size)	//stack full?
	{
		rv =1;
	}
	else
	{
		pS->data[pS->stactpointer++] = pE;
		pS->nElem++;	//increment number of elements
	}
	return rv;
}

//-----------------------------------------------
// Initialize function
//-----------------------------------------------

void InitMSplayer(void)
{
	pRepeatStack = newStack(64);	//allow nesting of up to 64 repeats
}
