//-----------------------------------
// Music Studio File Reader/Player
// defines
//
// Music Studio is an ancient Midi
// program that came out for the ST
// in the Mid 80's....sometime
// It was really a "toy" program,
// but you could actually do some
// very wonderful things with it.
// I really liked this program.
//
// Well, times moved on.  I started
// using PC's and then Windows.  The
// ST was soon gathering dust.  My
// ST now no longer functions. 
// Recently (Feb 2011) I ran across
// the floopies that had the back ups
// for my music studio songs.
// I managed to import them into my 
// win XP machine and decided I would
// write a "clone" of music studio
// that ran on Windows.
//
// File created by Jim Patchell
// The Blind Hacker
//----------------------------------

#ifndef MSPLAY__H
#define MSPLAY__H

/*
	Music Studio TOKENS
*/
#define EVENT_END	0x0
#define KEY_SIGNATURE	0x80
#define	TEMPO	0x81
#define	BAR	0x82
#define	TIME_SIGNATURE	0x83
#define	LOUDNESS	0x84
#define	REPEAT_START	0x85
#define	REPEAT_STOP	0x86
#define END	0xFF

#define CMAJ	0x01
#define	GMAJ	0x02
#define	DMAJ	0x03
#define	AMAJ	0x04
#define	EMAJ	0x05
#define	BMAJ	0x06
#define	FSMAJ	0x07
#define	CSMAJ	0x08
#define	FMAJ	0x09
#define	BFMAJ	0x0a
#define	EFMAJ	0x0b
#define	AFMAJ	0x0c
#define	DFMAJ	0x0d
#define	GFMAJ	0x0e
#define	CFMAJ	0x0f

//---------------------------
// defines for note tokens
//---------------------------

// BYTE 0
								//bit 0 ?
								//bit 1 ?
								//bit 2 ?
								//bit 3 ?
#define MS_REST_FLAG	(0x10)	//bit 4
#define MS_ENDTIE_FLAG	(0x20)	//bit 5
#define MS_BEGTIE_FLAG	(0x40)	//bit 6
								//bit 7 not used

// BYTE 1

#define SHARP	0x80
#define	FLAT	0xc0
#define	INKEY	0x00
#define	NATURAL	0x40
#define MS_KEY_MASK	(0xc0)
#define MS_ACCENT_FLAG	(0x20)
#define MS_DURATION_MASK	(0x1f)

// BYTE 2




typedef struct {
	char Name[10];
} Chan;

typedef struct {
	Chan Channel[15];		//channel names
	char chan_range[15];	/*	array contains data that coresponds to channel and range	*/
	char patch[15];			//	array that indicates midi patch for instrument
	long lyrics[5];
	char name[32];
}MIDI_INFO;

typedef struct {
	short int Channel;	//midi channel to send out data on;
	short int Duration;	//duration of event
	short int Note;		//note to be played
	short int Velocity;	//loudenss of note
	short int Track;	//Track note was on
	char RestFlag;
	char TieBegFlag;
	char TieEndFlag;
	char EventFree;		//true if event is not being used
}EVENT;

typedef struct {
	char name[11];
	char EnableFlag;
	short int range;
	short int patch;
	short int velocity;
	short int channel;
	short int octave;
	short int octaveNum;
}TRACK_INFO;

//song class
typedef struct {
	char SongName[34];
	TRACK_INFO	trkinfo[15];
	EVENT *pEvents;
	int nEvents;	//maximum number of events
	int Event;			//event counter
	char *Song;			//points to first byte of the song
	char *pCurSong;		//points to current position in song
	char SongData[1];	//this is where all the raw song data is
}MSSONG;

//---------------------------------
// Repeat Stack
//---------------------------------

typedef struct {
	int Count;	//repeat counter
	char *pRstart;	//pointer in song data to return to
}REPEAT;

//------------------------------
// REPEAT functions
//------------------------------

REPEAT *newRepeat(int cnt, char *pRstrt);
void deleteRepeat(REPEAT *pR);
int DecRepeatCount(REPEAT *pR);

//-----------------------------
// Stack type
//-----------------------------

typedef struct {
	int stactpointer;	//stack pointer (duh)
	int nElem;			//number of data elements on stack
	int Size;			//max size of stack
	void *data[1];	//array of pointers to data elements
} STACK;

//-----------------
// stack functions
//-----------------

STACK *newStack(int size);
void *GetStackTop(STACK *pS);
void *PopStack(STACK *pS);
int PushStack(STACK *pS,void *pE);
//--------------------------------
// Song File Offsets
//--------------------------------

#define MS_OFFSET_MIDI_INFO		(0x118)
#define MS_OFFSET_SONG_DATA		(0x200)
#define MS_OFFSET_MUSIC_DATA	(0x208)

//--------------------------------
// misc defines
//--------------------------------

#define FALSE	(0)
#define TRUE	(1)

//--------------------------------
// Function Prototypes
//--------------------------------

extern void StartSong(MSSONG *);
extern int  parse(MSSONG *);
extern EVENT *search(int note,int chan,EVENT *pEv, int MaxEvents);
extern EVENT *GetFreeEvent(EVENT *pEv,int MaxEvents);
extern void WaitAllOff(MSSONG *pSong);
extern void timing(MSSONG *pSong,int n_duration);
extern void midi_off(MSSONG *pSong);
extern int ticker(MSSONG *pSong);
extern void do_stop(void);
extern void InitBuffer(MSSONG *pSong);
extern MSSONG *LoadFile(char *SongName,int nEvents);
extern MSSONG * LoadNextSong(void);
extern void GetSongParameters(MSSONG *pSong);

extern int sPrintTrkInfo(char *s,MSSONG *pSong);

extern void InitMSplayer(void);

#endif
